"""Module level object."""
from .__version__ import __version__
from .__version__ import __description__
from .__version__ import __url__
from .__version__ import __title__
from .__version__ import __author__
from .__version__ import __author_email__
from .__version__ import __license__
from .__version__ import __copyright__


__all__ = (
    __version__,
    __description__,
    __url__,
    __title__,
    __author__,
    __author_email__,
    __license__,
    __copyright__,
)



"""Telemetry events."""
import json

from chicken_dinner.models.telemetry.objects import TelemetryObject
from chicken_dinner.util import camel_to_snake
from chicken_dinner.util import remove_from_dict


class TelemetryEvent(object):
    """Telemetry event model.

    Generic object for all Telemetry events. To identify the event type use the
    object attribute ``event_type``.

    Provides an object attribute based model for telemetry events, creating
    embedded telemetry objects recursively. Converts all event and object keys
    to snake-cased key names.

    :param data: the JSON object data associated with the telemetry event
    """

    def __init__(self, data):
        for k, v in data.items():
            if isinstance(v, dict):
                setattr(self, camel_to_snake(k), TelemetryObject(v, k))
            elif isinstance(v, list):
                setattr(self, camel_to_snake(k), [TelemetryObject(e, k) for e in v])
            elif k in ("_D", "_T", "_V"):
                setattr(self, k, v)
            elif k == "blueZoneCustomOptions":  # serialized json
                setattr(self, camel_to_snake(k), [TelemetryObject(e, k) for e in json.loads(v)])
            else:
                setattr(self, camel_to_snake(k), v)

    @property
    def event_type(self):
        """The snake cased event type from _T."""
        return camel_to_snake(self._T)

    @property
    def timestamp(self):
        """The timestamp value from _D."""
        return self._D

    def __getitem__(self, key):
        return getattr(self, camel_to_snake(key))

    def __str__(self):
        return "TelemetryEvent for " + self.event_type

    def __repr__(self):
        return "TelemetryEvent({d})".format(d=self.dumps())

    def dumps(self):
        """Serialize the event to a JSON string."""
        return json.dumps(
            self,
            default=lambda x: remove_from_dict(x.__dict__, ["reference"]),
            sort_keys=True,
            indent=4
        )

    def to_dict(self):
        """Get the event as a dict."""
        return json.loads(self.dumps())



"""Match asset model."""


class Asset(object):


    def __init__(self, pubg, match, data, shard=None):
        self._pubg = pubg
        self._shard = shard
        #: The Match object for this asset
        self.match = match
        #: The data payload associated with this asset
        self.data = data

    @property
    def shard(self):
        """The shard for the match associated with this asset."""
        return self._shard or self._pubg.shard

    @property
    def id(self):
        """The asset id."""
        return self.data["id"]

    @property
    def description(self):
        """A description of the asset."""
        return self.data["attributes"]["description"]

    @property
    def created_at(self):
        """When the asset was created."""
        return self.data["attributes"]["createdAt"]

    @property
    def url(self):
        """The URL of the asset."""
        return self.data["attributes"]["URL"]



"""Telemetry objects."""
import json

from chicken_dinner.util import camel_to_snake
from chicken_dinner.util import remove_from_dict


class TelemetryObject(object):
    """Telemetry object model.

    Generic object for telemetry event objects.

    Provides an object-attribute based model for telemetry objects, creating
    embedded telemetry objects recursively. Converts all event and object keys
    to snake-cased key names.

    :param data: the JSON object data associated with the telemetry object
    :param reference: the key from the parent object that refernces this object
    """

    def __init__(self, data, reference):
        #: The key name that references this object
        self.reference = camel_to_snake(reference)
        for k, v in data.items():
            if isinstance(v, dict):
                setattr(self, camel_to_snake(k), TelemetryObject(v, k))
            elif isinstance(v, list) and len(v) > 0 and isinstance(v[0], dict):
                setattr(self, camel_to_snake(k), [TelemetryObject(e, k) for e in v])
            else:
                setattr(self, camel_to_snake(k), v)

    def __getitem__(self, key):
        return getattr(self, camel_to_snake(key))

    def __str__(self):
        return "TelemetryObject " + self.reference + " object"

    def __repr__(self):
        return "TelemetryObject({d}, {r})".format(d=self.dumps(), r=self.reference)

    def dumps(self):
        """Serialize the event object to a JSON string."""
        return json.dumps(
            self,
            default=lambda x: remove_from_dict(x.__dict__, ["reference"]),
            sort_keys=True,
            indent=4
        )

    def to_dict(self):
        """Get the event object as a dict."""
        return json.loads(self.dumps())



"""Tournaments model."""
from chicken_dinner.models.tournament import Tournament


class Tournaments(object):
    """Tournaments model.

    An object encapsulating metadata about a set of PUBG tournaments.

    :param pubg: an instance of the class :class:`chicken_dinner.pubgapi.PUBG`
    """

    def __init__(self, pubg, shard="pc-tournament"):
        self._pubg = pubg
        self._shard = shard
        #: The API response for this object.
        self.response = self._pubg._core.tournaments()
        self._tournaments = [
            Tournament(self._pubg, t["id"], t["attributes"]["createdAt"], self.shard)
            for t in self.response["data"]
        ]

    def __getitem__(self, idx):
        return self._tournaments[idx]

    @property
    def data(self):
        """The data payload of the repsonse."""
        return self.response["data"]

    @property
    def ids(self):
        """The tournament ids."""
        return [
            t["id"] for t in self.data
        ]

    @property
    def meta(self):
        """The meta information attached to the response."""
        return self.response["meta"]

    @property
    def shard(self):
        """The shard for these tournaments."""
        return self._shard

    @property
    def url(self):
        """The URL for this tournaments resource."""
        return self.response["links"]["self"]



"""Tournament model."""
from chicken_dinner.models.match import Match


class Tournament(object):
    """Tournament model.

    An object encapsulating metadata about a PUBG tournament.

    :param pubg: an instance of the class :class:`chicken_dinner.pubgapi.PUBG`
    :param str tournament_id: a tournament id
    """

    def __init__(self, pubg, tournament_id, created_at=None, shard="pc-tournament"):
        self._pubg = pubg
        self._shard = shard
        self._id = tournament_id
        self._created_at = created_at
        self._response = None

    @property
    def created_at(self):
        """The time at which this tournament resource was created."""
        return self._created_at

    @property
    def data(self):
        """The data payload of the response."""
        return self.response["data"]

    @property
    def response(self):
        """The response for this tournament resource."""
        if self._response is None:
            self._response = self._pubg._core.tournament(self.id)
        return self._response

    @property
    def id(self):
        """The tournament id."""
        return self._id

    @property
    def match_ids(self):
        """The match ids associated with this tournament."""
        return [
            m["id"] for m in self.response["included"]
        ]

    def get_matches(self):
        """Get a list of match objects for the tournament matches."""
        return [
            Match(self._pubg, match_id, shard=self.shard) for match_id in self.match_ids
        ]

    @property
    def meta(self):
        """The meta information attached to the response."""
        return self.response["meta"]

    @property
    def shard(self):
        """The shard for these tournaments."""
        return self._shard

    @property
    def url(self):
        """The URL for this tournaments resource."""
        return self.response["links"]["self"]



"""Single season model."""
import chicken_dinner.models.player


class Season(object):
    """Single season model.

    An object containing metadata about a PUBG season.

    :param pubg: an instance of the class :class:`chicken_dinner.pubgapi.PUBG`
    :param data: the ``data`` payload from a season response
    :param shard: the shard for the season response
    """

    def __init__(self, pubg, data, shard=None):
        self._pubg = pubg
        self._shard = shard
        #: The data payload for this season
        self.data = data

    @property
    def shard(self):
        """The shard for this season."""
        return self._shard or self._pubg.shard

    @property
    def id(self):
        """The season id."""
        return self.data["id"]

    def is_current(self):
        """Return True if this season is the current season."""
        return self.data["attributes"]["isCurrentSeason"]

    def is_offseason(self):
        """Return True if this season is in the offseason."""
        return self.data["attributes"]["isOffseason"]

    def get_player(self, player_id):
        """Get a player-season dataset for this season.

        :param player_id: a player account id or instance of
            :class:`chicken_dinner.models.Player` to retrieve data for this
            season
        :return: a :class:`chicken_dinner.models.PlayerSeason` for this
            ``player_id``
        """
        if isinstance(player_id, chicken_dinner.models.player.Player):
            player_id = player_id.id
        return self._pubg.player_season(player_id, self.id, self.shard)



"""Match model."""
from .asset import Asset
from .participant import Participant
from .roster import Roster
from chicken_dinner.constants import game_mode_to_gp
from chicken_dinner.constants import map_to_map_name
from chicken_dinner.models.telemetry import Telemetry


class Match(object):
    """Match object.

    :param pubg: a PUBG instance
    :param str match_id: the ``match_id`` for this match
    :param str shard: the shard for this match
    """

    def __init__(self, pubg, match_id, shard=None):
        self._pubg = pubg
        self._shard = shard
        #: The match id for this match
        self.match_id = match_id
        #: The API response for this object
        self.response = self._pubg._core.match(match_id, shard)
        self.roster_to_participant = {}
        self.participant_to_roster = {}
        self._participant_data = {}
        rosters_idx = []
        for idx, item in enumerate(self.response["included"]):
            if item["type"] == "roster":
                rosters_idx.append(idx)
                self.roster_to_participant[item["id"]] = []
                for participant in item["relationships"]["participants"]["data"]:
                    self.roster_to_participant[item["id"]].append(participant["id"])
                    self.participant_to_roster[participant["id"]] = item["id"]
            elif item["type"] == "participant":
                self._participant_data[item["id"]] = item
            elif item["type"] == "asset":
                self.asset = Asset(pubg, self, item, shard)

        #: A list of Roster instances for this match
        self.rosters = [
            Roster(pubg, self, self.response["included"][idx], shard)
            for idx in rosters_idx
        ]

    @property
    def shard(self):
        """The shard for this match."""
        return self._shard or self._pubg.shard

    @property
    def asset_id(self):
        """The asset id for this match."""
        return self.asset.id

    @property
    def data(self):
        """The data payload from the response for this match."""
        return self.response["data"]

    @property
    def created_at(self):
        """When the match was created."""
        return self.data["attributes"]["createdAt"]

    @property
    def duration(self):
        """The duration of the match."""
        return self.data["attributes"]["duration"]

    @property
    def game_mode(self):
        """The game mode for the match."""
        return game_mode_to_gp[self.data["attributes"]["gameMode"]]

    @property
    def id(self):
        """The match id."""
        return self.data["id"]

    @property
    def map_name(self):
        """The name of the map on which this match was played."""
        return map_to_map_name[self.data["attributes"]["mapName"]]

    @property
    def map_id(self):
        """The map id on which this match was played."""
        return self.data["attributes"]["mapName"]

    @property
    def participants(self):
        """A list of Participant instances for match participants."""
        return [
            participant for roster in self.rosters
            for participant in roster.participants
        ]

    @property
    def stats(self):
        """Stats for this match."""
        return self.data["attributes"]["stats"]

    @property
    def tags(self):
        """Tags associated with this match."""
        return self.data["attributes"]["tags"]

    @property
    def title_id(self):
        """The title id associated with this match."""
        return self.data["attributes"]["titleId"]

    @property
    def url(self):
        """The URL for this match resource."""
        return self.data["links"]["self"]

    @property
    def telemetry_url(self):
        """The URL for the telemetry data for this match."""
        return self.asset.url

    def get_telemetry(self):
        """Download match telemetry and create a Telemetry object instance.

        :return: a :class:`chicken_dinner.models.telemetry.Telemetry` instance
            for this match.
        """
        return Telemetry(self._pubg, self.telemetry_url, shard=self.shard)

    @property
    def rosters_player_names(self):
        """A mapping of roster_ids to player names for this match."""
        return {
            roster.id: roster.player_names for roster in self.rosters
        }

    @property
    def winner(self):
        """The Roster instance for the winner of this match."""
        for roster in self.rosters:
            if roster.won == True:
                return roster

    @property
    def is_custom(self):
        """Whether or not the match is a custom match."""
        return self.data["attributes"]["isCustomMatch"]



# flake8: noqa
from chicken_dinner.models.match import Match
from chicken_dinner.models import Player
from chicken_dinner.models import Players
from chicken_dinner.models import PlayerSeason
from chicken_dinner.models import Samples
from chicken_dinner.models import Seasons
from chicken_dinner.models import Status
from chicken_dinner.models import Tournament
from chicken_dinner.models import Tournaments
from chicken_dinner.pubgapi.core import PUBGCore
from chicken_dinner.pubgapi.rate_limiter import DEFAULT_CALL_COUNT
from chicken_dinner.pubgapi.rate_limiter import DEFAULT_CALL_WINDOW


class PUBG(object):
    """High level type-based interface to the PUBG JSON API.

    :param str api_key: your PUBG api key
    :param str shard: (optional) a shard to use for all requests with this
        instance
    :param bool gzip: (default=*True*) whether to gzip the responses. Responses
        are automatically unzipped by the underlying ``requests`` library.
    :param int limit_call_count: (optional) your api key rate limit count
    :param int limit_call_window: (optional) your api key rate limit window
    """

    def __init__(self, api_key, shard=None, gzip=True,
                 limit_call_count=DEFAULT_CALL_COUNT,
                 limit_call_window=DEFAULT_CALL_WINDOW):
        self._core = PUBGCore(
            api_key, shard, gzip, limit_call_count, limit_call_window
        )

    @property
    def api_key(self):
        return self._core.api_key

    @api_key.setter
    def api_key(self, value):
        self._core.api_key = value

    @property
    def shard(self):
        return self._core.shard

    @shard.setter
    def shard(self, value):
        self._core.shard = value

    def current_season(self, shard=None):
        """Get the current season.

        :param str shard: (optional) the shard to use if different from the
            instance shard
        :return: a :class:`chicken_dinner.models.Season` object for the current
            season
        """
        shard = shard or self.shard
        return Seasons(self, shard).current()

    def match(self, match_id, shard=None):
        """Get match info by ``match_id``.

        :param str match_id: the match_id to query
        :param str shard: (optional) the shard to use if different from the
            instance shard
        :return: a :class:`chicken_dinner.models.Match` object for the match
            corresponding to ``match_id``.
        """
        shard = shard or self.shard
        return Match(self, match_id, shard)

    def samples(self, shard=None):
        """Get match samples.

        :param str shard: (optional) the shard to use if different from the
            instance shard
        :return: a :class:`chicken_dinner.models.Samples` object containing
            match samples.
        """
        shard = shard or self.shard
        return Samples(self, shard)

    def seasons(self, shard=None):
        """Get an iterable of PUBG seasons.

        :param str shard: (optional) the shard to use if different from the
            instance shard
        :return: an iterable :class:`chicken_dinner.models.Seasons` object
            containing multiple :class:`chicken_dinner.models.Season` objects.
        """
        shard = shard or self.shard
        return Seasons(self, shard)

    def status(self):
        """Get the status of the PUBG API.

        :return: a :class:`chicken_dinner.models.Status` object containing
            information about the status of the API.
        """
        return Status(self)

    def player_season(self, player_id, season_id, shard=None):
        """Get a player's information for a particular season.

        :param str player_id: the player's account id
        :param str season_id: the PUBG season id or "current" for current
            season
        :param str shard: (optional) the shard to use if different from the
            instance shard
        :return: a :class:`chicken_dinner.models.PlayerSeason` object
            containing information about matches and statistics for a player's
            performance in the season given by ``season_id``
        """
        shard = shard or self.shard
        if season_id == "current":
            season_id = self.current_season(shard).id
        if isinstance(player_id, Player):
            player_id = player_id.id
        return PlayerSeason(self, player_id, season_id, shard)

    def player(self, player_id, shard=None):
        """Get a player's metadata.

        :param str player_id: the player's account id
        :param str shard: (optional) the shard to use if different from the
            instance shard
        :return: a :class:`chicken_dinner.models.Player` object
            containing information about the player
        """
        shard = shard or self.shard
        return Player(self, player_id, shard)

    def players(self, filter_type, filter_value, shard=None):
        """Get multiple players' metadata.

        :param str filter_type: query by either "player_ids" or "player_names"
        :param list filter_value: a list of strings of the ``player_ids`` or
            ``player_names`` to search
        :param str shard: (optional) the shard to use if different from the
            instance shard
        :return: an iterable :class:`chicken_dinner.models.Players` object
            consisting of multiple :class:`chicken_dinner.models.Player`
            objects
        """
        shard = shard or self.shard
        return Players(self, filter_type, filter_value, shard)

    def players_from_ids(self, player_ids, shard=None):
        """Get multiple players' metadata from a list of ``player_ids``.

        :param list player_ids: a list of strings of ``player_ids``
        :param str shard: (optional) the shard to use if different from the
            instance shard
        :return: an iterable :class:`chicken_dinner.models.Players` object
            consisting of multiple :class:`chicken_dinner.models.Player`
            objects
        """
        shard = shard or self.shard
        return Players(self, "player_ids", player_ids, shard)

    def players_from_names(self, player_names, shard=None):
        """Get multiple players' metadata from a list of ``player_names``.

        :param list player_names: a list of strings of ``player_names``
        :param str shard: (optional) the shard to use if different from the
            instance shard
        :return: an iterable :class:`chicken_dinner.models.Players` object
            consisting of multiple :class:`chicken_dinner.models.Player`
            objects
        """
        shard = shard or self.shard
        return Players(self, "player_names", player_names, shard)

    def telemetry(self, url):
        """Get a telemetry object from a telemetry url.

        :param str url: the url for the telemetry data
        :return: a :class:`chicken_dinner.models.telemetry.Telemetry` object
        """
        return Telemetry(self, url)

    def tournament(self, tournament_id):
        """Get a tournament by its id.

        :param str tournament_id: the tournament id for which to get data
        :return: a :class:`chicken_dinner.models.Tournament` object
        """
        return Tournament(self, tournament_id)

    def tournaments(self):
        """Get a list of tournaments.

        :return: a :class:`chicken_dinner.models.Tournaments` object
        """
        return Tournaments(self)



"""Rate limiter class."""
from collections import deque
import logging
import time


DEFAULT_CALL_COUNT = 10
DEFAULT_CALL_WINDOW = 60


class RateLimiter(object):
    """The RateLimiter class.

    This class stores a specific rate limit call/window pair as well as the
    timestamps of recent calls within the window. It can check against the
    limit and add calls to the deque of calls.
    """

    # A fixed time in seconds to add to all sleeps as a safety buffer.
    buffer_time = 2

    def __init__(self, calls=DEFAULT_CALL_COUNT, window=DEFAULT_CALL_WINDOW):
        """Instantiate the class."""
        self.calls = calls
        self.window = window
        self.deq = deque()

    def check(self):
        """Check the previous calls and discard calls ouside of the window."""
        if len(self.deq) >= self.calls:
            # Check calls history
            for k in range(len(self.deq)):
                # Trash the calls that are before the start of the window
                if (len(self.deq) > 0 and
                        self.deq[0] < time.time() - self.window):
                    self.deq.popleft()
                else:
                    break
        # If at the limit still, store the required wait time.
        if len(self.deq) >= self.calls:
            return self.deq[0] - (time.time() - self.window) + self.buffer_time
        else:
            return 0

    def call(self, check=True):
        """Check the deque, optionally sleep, and add a call to the window."""
        if check:
            sleep_seconds = self.check()
            if sleep_seconds > 0:
                logging.info("Rate limiter sleeping for {} seconds.".format(
                    sleep_seconds
                ))
                time.sleep(self.check())
        self.deq.append(time.time())



"""Constants and reference maps from pythonic names to pubg api names."""


PERSPECTIVES = ("tpp", "fpp")
GROUPS = ("solo", "duo", "squad")

gp_to_game_mode = {
    "solo": "solo",
    "duo": "duo",
    "squad": "squad",
    "solo-fpp": "solo-fpp",
    "duo-fpp": "duo-fpp",
    "squad-fpp": "squad-fpp",
}
game_mode_to_gp = {
    "solo": "solo",
    "duo": "duo",
    "squad": "squad",
    "solo-fpp": "solo-fpp",
    "duo-fpp": "duo-fpp",
    "squad-fpp": "squad-fpp",
}
gp_to_matches = {
    "solo": "matchesSolo",
    "duo": "matchesDuo",
    "squad": "matchesSquad",
    "solo-fpp": "matchesSoloFPP",
    "duo-fpp": "matchesDuoFPP",
    "squad-fpp": "matchesSquadFPP",
}
matches_to_gp = {
    "matchesSolo": "solo",
    "matchesDuo": "duo",
    "matchesSquad": "squad",
    "matchesSoloFPP": "solo-fpp",
    "matchesDuoFPP": "duo-fpp",
    "matchesSquadFPP": "squad-fpp",
}

map_to_map_name = {
    "Desert_Main": "Miramar",
    "Erangel_Main": "Erangel",
    "Savage_Main": "Sanhok",
}

map_name_to_map = {
    "Miramar": "Desert_Main",
    "Erangel": "Erangel_Main",
    "Sanhok": "Savage_Main",
}

map_dimensions = {
    "Desert_Main": [819200, 819200],
    "Erangel_Main": [819200, 819200],
    "Savage_Main": [409600, 409600],
}

true_false = {
    "false": False,
    "true": True,
}

BASE_URL = "https://api.playbattlegrounds.com"
SHARD_URL = BASE_URL + "/shards/"
STATUS_URL = BASE_URL + "/status"
TOURNAMENTS_URL = BASE_URL + "/tournaments"
PLAYER_FILTERS = {
    "player_ids": "playerIds",
    "player_names": "playerNames",
}
SHARDS = [
    "xbox-as",
    "xbox-eu",
    "xbox-na",
    "xbox-oc",
    "pc-krjp",
    "pc-jp",
    "pc-na",
    "pc-eu",
    "pc-oc",
    "pc-kakao",
    "pc-sea",
    "pc-sa",
    "pc-as",
    "pc-tournament",
]

COLORS = [
    "#fc3f3f",
    "#d93636",
    "#eb7575",
    "#eb583b",
    "#fc937e",
    "#d97e6c",
    "#eb753b",
    "#fca87e",
    "#d9916c",
    "#eb933b",
    "#ebb075",
    "#fcbd3f",
    "#d9a336",
    "#ebc475",
    "#fcdd3f",
    "#d9be36",
    "#fce77e",
    "#d9c76c",
    "#ebeb3b",
    "#fcfc7e",
    "#d9d96c",
    "#ddfc3f",
    "#bed936",
    "#d7eb75",
    "#bdfc3f",
    "#a3d936",
    "#d2fc7e",
    "#b5d96c",
    "#9efc3f",
    "#87d936",
    "#bdfc7e",
    "#a3d96c",
    "#7efc3f",
    "#6cd936",
    "#a8fc7e",
    "#91d96c",
    "#58eb3b",
    "#89eb75",
    "#3ffc3f",
    "#7efc7e",
    "#6cd96c",
    "#3beb58",
    "#7efc93",
    "#6cd97e",
    "#3ffc7e",
    "#36d96c",
    "#7efca8",
    "#6cd991",
    "#3beb93",
    "#7efcbd",
    "#6cd9a3",
    "#3ffcbd",
    "#36d9a3",
    "#75ebc4",
    "#3bebcd",
    "#7efce7",
    "#6cd9c7",
    "#3ffcfc",
    "#36d9d9",
    "#7efcfc",
    "#6cd9d9",
    "#3fddfc",
    "#36bed9",
    "#7ee7fc",
    "#6cc7d9",
    "#3fbdfc",
    "#36a3d9",
    "#7ed2fc",
    "#6cb5d9",
    "#3f9efc",
    "#3687d9",
    "#7ebdfc",
    "#6ca3d9",
    "#3f7efc",
    "#366cd9",
    "#759ceb",
    "#3b58eb",
    "#7e93fc",
    "#6c7ed9",
    "#3636d9",
    "#7e7efc",
    "#5f3ffc",
    "#937efc",
    "#7e6cd9",
    "#7e3ffc",
    "#6c36d9",
    "#9c75eb",
    "#9e3ffc",
    "#8736d9",
    "#bd7efc",
    "#a36cd9",
    "#bd3ffc",
    "#a336d9",
    "#d27efc",
    "#b56cd9",
    "#dd3ffc",
    "#be36d9",
    "#e77efc",
    "#c76cd9",
    "#fc3ffc",
    "#d936d9",
    "#fc7efc",
    "#d96cd9",
    "#eb3bcd",
    "#fc7ee7",
    "#d96cc7",
    "#fc3fbd",
    "#d936a3",
    "#fc7ed2",
    "#d96cb5",
    "#eb3b93",
    "#eb75b0",
    "#fc3f7e",
    "#d9366c",
    "#eb759c",
    "#fc3f5f",
    "#d93651",
    "#fc7e93",
    "#d96c7e",
]



# flake8: noqa
import requests

from chicken_dinner.pubgapi.rate_limiter import DEFAULT_CALL_COUNT
from chicken_dinner.pubgapi.rate_limiter import DEFAULT_CALL_WINDOW
from chicken_dinner.pubgapi.rate_limiter import RateLimiter
from chicken_dinner.constants import BASE_URL
from chicken_dinner.constants import SHARD_URL
from chicken_dinner.constants import STATUS_URL
from chicken_dinner.constants import TOURNAMENTS_URL
from chicken_dinner.constants import SHARDS
from chicken_dinner.constants import PLAYER_FILTERS


class PUBGCore(object):
    """Low level interface to the PUBG JSON API.

    Provides methods for interfacing directly with the PUBG JSON API. Returns
    deserialized JSON responses.

    Info: https://documentation.playbattlegrounds.com/en/introduction.html

    :param str api_key: your PUBG api key
    :param str shard: (optional) the shard to use in all requests for this
        instance
    :param bool gzip: (optional) compress responses as gzip
    :param int limit_call_count: (optional) your api key rate limit count
    :param int limit_call_window: (optional) your api key rate limit window
    """

    def __init__(self, api_key, shard=None, gzip=True,
                 limit_call_count=DEFAULT_CALL_COUNT,
                 limit_call_window=DEFAULT_CALL_WINDOW):
        self.session = requests.Session()
        self.api_key = api_key
        if gzip:
            self.session.headers.update({
                "Accept-Encoding": "gzip",
            })
        self.rate_limiter = RateLimiter(limit_call_count, limit_call_window)
        if shard is None or shard in SHARDS:
            self.shard = shard
        else:
            raise ValueError("Invalid shard provided.")

    @property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, value):
        self._api_key = value
        self.session.headers = {
            "Authorization": "Bearer " + value,
            "Accept": "application/vnd.api+json",
        }

    def _check_shard(self, shard):
        shard = shard or self.shard
        if shard is None:
            raise ValueError("A shard must be provided.")
        elif shard not in SHARDS:
            raise ValueError("Invalid shard provided.")
        else:
            return shard

    def _get(self, url, params=None, limited=True):
        if limited and self.rate_limiter.window > 0:
            self.rate_limiter.call()
        response = self.session.get(url, params=params)
        return response

    def match(self, match_id, shard=None):
        """Get a response from the match endpoint.

        Description: https://documentation.playbattlegrounds.com/en/matches-endpoint.html

        Calls here do not apply to the rate limit.

        :param str match_id: the ``match_id`` to query
        :param str shard: (optional) the ``shard`` to use if different from
            the one used on instantiation
        :return: the json response from ``/{shard}/matches/{match_id}``
        """
        shard = self._check_shard(shard)
        url = SHARD_URL + shard + "/matches/" + match_id
        return self._get(url, limited=False).json()

    def player(self, player_id, shard=None):
        """Get a response from the player endpoint.

        Endpoints: https://documentation.playbattlegrounds.com/en/players-endpoint.html

        :param str player_id: the PUBG ``player_id`` (account id) to query
        :param str shard: (optional) the ``shard`` to use if different from
            the one used on instantiation
        :return: the JSON response from ``/{shard}/players/{player_id}``
        """
        shard = self._check_shard(shard)
        url = SHARD_URL + shard + "/players/" + str(player_id)
        return self._get(url).json()

    def player_season(self, player_id, season_id, shard=None):
        """Get a response from the player/season endpoint.

        Endpoints: https://documentation.playbattlegrounds.com/en/players-endpoint.html

        :param str player_id: the PUBG ``player_id`` (account id) to query
        :param str season: the ``season_id`` to query
        :param str shard: (optional) the ``shard`` to use if different from
            the one used on instantiation
        :return: the JSON response from
            ``/{shard}/players/{player_id}/seasons/{season_id}``
        """
        shard = self._check_shard(shard)
        url = SHARD_URL + shard + "/players/" + str(player_id)
        url = url + "/seasons/" + str(season_id)
        return self._get(url).json()

    def players(self, filter_type, filter_value, shard=None):
        """Get a response from the players endpoint.

        Description: https://documentation.playbattlegrounds.com/en/players-endpoint.html

        :param str filter_type: query by either "player_ids" or "player_names"
        :param list filter_value: a list of strings of the ``player_ids`` or
            ``player_names`` to search
        :param str shard: (optional) the ``shard`` to use if different from
            the one used on instantiation
        :return: the response from the ``/{shard}/players`` endpoint
        """
        shard = self._check_shard(shard)
        if filter_type not in PLAYER_FILTERS:
            raise ValueError("Filter type must be in " + str(PLAYER_FILTERS))
        if isinstance(filter_value, list):
            filter_value = ",".join(filter_value)

        params = {"filter[" + PLAYER_FILTERS[filter_type] + "]": filter_value}
        url = SHARD_URL + shard + "/players"
        return self._get(url, params).json()

    def samples(self, start=None, shard=None):
        """Get a response from the samples endpoint.

        Description: https://documentation.playbattlegrounds.com/en/samples-endpoint.html

        :param str start: (optional) the start timestamp from which to get
            samples
        :param str shard: (optional) the ``shard`` to use if different from
            the one used on instantiation
        :return: the JSON response from the ``/{shard}/samples`` endpoint
        """
        shard = self._check_shard(shard)
        url = SHARD_URL + shard + "/samples"
        params = {}
        if start is not None:
            params = {"filter[createdAt-start]": start}
        return self._get(url, params).json()

    def seasons(self, shard=None):
        """Get a response from the seasons endpoint.

        Description: https://documentation.playbattlegrounds.com/en/players-endpoint.html#/Seasons/get_seasons

        :param str shard: (optional) the ``shard`` to use if different from
            the one used on instantiation
        :return: the JSON response from the ``/{shard}/seasons`` endpoint.
        """
        shard = self._check_shard(shard)
        url = SHARD_URL + shard + "/seasons"
        return self._get(url).json()

    def status(self):
        """Get a response from the status endpoint.

        Description: https://documentation.playbattlegrounds.com/en/status-endpoint.html

        :return: the JSON response from the ``/status`` endpoint.
        """
        return self._get(STATUS_URL).json()

    def telemetry(self, url):
        """Download the telemetry data.

        Description: https://documentation.playbattlegrounds.com/en/telemetry.html

        Calls here do not apply to the rate limit.

        :param str url: the telemetry data URL
        :return: the JSON response for the telemetry URL
        """
        return self._get(url, limited=False).json()

    def tournament(self, tournament_id):
        """Get information about a tournament.

        Description: https://documentation.playbattlegrounds.com/en/tournaments-endpoint.html#/Tournaments/get_tournaments__id_

        :param str tournament_id: the tournament ID on which to retrieve data
        :return: the JSON response for the tournament id
        """
        return self._get(TOURNAMENTS_URL + "/" + tournament_id).json()

    def tournaments(self):
        """Get a list of tournaments.

        Description: https://documentation.playbattlegrounds.com/en/tournaments-endpoint.html#/Tournaments/get_tournaments

        :return: the JSON response for the tournaments endpoint
        """
        return self._get(TOURNAMENTS_URL).json()



"""Telemetry class."""
import datetime

from chicken_dinner.constants import map_to_map_name
from chicken_dinner.models.telemetry.events import TelemetryEvent


class Telemetry(object):
    """Telemetry model.

    :param pubg: a PUBG instance
    :param str shard: the shard for the match associated with this telemetry
    :param str url: the url for this telemetry
    :param list telemetry_json: (optional) the raw telemetry response
    """

    def __init__(self, pubg, url, telemetry_json=None, shard=None):
        self._pubg = pubg
        self._shard = shard
        if telemetry_json is not None:
            #: The API response associated with this object
            self.response = telemetry_json
        else:
            self.response = self._pubg._core.telemetry(url)
        #: Snake cased object-attribute models for telemetry events and objects
        self.events = [TelemetryEvent(e) for e in self.response]
        if getattr(self.events[-1], "common", None) is not None:
            #: The platform for this game, "pc" or "xbox"
            self.platform = "pc"
        else:
            self.platform = "xbox"
        #: Whether this game was played on PC
        self.is_pc = "pc" == self.platform
        #: Whether this game was played on xbox
        self.is_xbox = "xbox" == self.platform

    def __getitem__(self, key):
        return self.events[key]

    @property
    def shard(self):
        """The shard for this match."""
        return self._shard or self._pubg.shard

    def event_types(self):
        """A sorted list of event type names from this telemetry."""
        return sorted(list(set([e.event_type for e in self.events])))

    def filter_by(self, event_type=None): #, account_id=None, player_name=None):
        """Get a list of telemetry events for a specific event type.

        :param event_type: the event type to filter
        """
        events = None
        if event_type is not None:
            events = [
                event for event in self.events
                if event.event_type == event_type
            ]
        else:
            events = [event for event in self.events]

        return events

    def player_ids(self):
        """The account ids of all players in the match."""
        accounts = []
        for event in self.events:
            if event.event_type == "log_player_login":
                accounts.append(event.account_id)
        return accounts

    def players(self):
        """A map of player names to account ids for all players this match."""
        players = {}
        for event in self.events:
            if event.event_type == "log_player_create":
                players[event.character.name] = event.character.account_id
        return players

    def player_names(self):
        """A list of player names for all match pariticipants."""
        player_names = []
        for event in self.events:
            if event.event_type == "log_player_create":
                player_names.append(event.character.name)
        return player_names

    def damage_done(self, player=None, combat_only=True, distribution=False):
        """Damage done by each player in the match.

        :param str player: a player name to filter on
        :param bool combat_only: return only PvP damage (default True)
        :param bool distribution: return a player to player distribution of
            damage for the match if true. if false return total damage done
            by each player. (default False)
        """
        start = datetime.datetime.strptime(
            self.filter_by("log_match_start")[0].timestamp,
            "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        damage = {}
        damage_events = self.filter_by("log_player_take_damage")
        for event in damage_events:
            timestamp = datetime.datetime.strptime(
                event.timestamp, "%Y-%m-%dT%H:%M:%S.%fZ"
            )
            dt = (timestamp - start).total_seconds()
            if dt < 0 or event.attack_id == -1:
                continue
            victim = event.victim.name
            try:
                attacker = event.attacker.name
            except AttributeError:
                if combat_only:
                    continue
                else:
                    attacker = "[" + event.damage_type_category + "]"
            if player is not None and player != attacker:
                continue
            if distribution:
                if attacker not in damage:
                    damage[attacker] = {}
                if victim not in damage[attacker]:
                    damage[attacker][victim] = event.damage
                else:
                    damage[attacker][victim] += event.damage
            else:
                if attacker not in damage:
                    damage[attacker] = event.damage
                else:
                    damage[attacker] += event.damage
        return damage

    def damage_taken(self, player=None, combat_only=True, distribution=False):
        """Damage taken by each player in the match.

        :param str player: a player name to filter on
        :param bool combat_only: return only PvP damage (default True)
        :param bool distribution: return a player to player distribution of
            damage for the match if true. if false return total damage taken
            by each player. (default False)
        """
        start = datetime.datetime.strptime(
            self.filter_by("log_match_start")[0].timestamp,
            "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        damage = {}
        damage_events = self.filter_by("log_player_take_damage")
        for event in damage_events:
            timestamp = datetime.datetime.strptime(
                event.timestamp, "%Y-%m-%dT%H:%M:%S.%fZ"
            )
            dt = (timestamp - start).total_seconds()
            if dt < 0 or event.attack_id == -1:
                continue
            victim = event.victim.name
            if player is not None and player != victim:
                continue
            try:
                attacker = event.attacker.name
            except AttributeError:
                if combat_only:
                    continue
                else:
                    attacker = "[" + event.damage_type_category + "]"
            if distribution:
                if victim not in damage:
                    damage[victim] = {}
                if attacker not in damage[victim]:
                    damage[victim][attacker] = event.damage
                else:
                    damage[victim][attacker] += event.damage
            else:
                if victim not in damage:
                    damage[victim] = event.damage
                else:
                    damage[victim] += event.damage
        return damage

    def rosters(self):
        """The team rosters for the match."""
        rosters = {}
        for event in self.events[::-1]:
            if event.event_type == "log_match_end":
                for player in event.characters:
                    team = player.team_id
                    player_name = player.name
                    if team not in rosters:
                        rosters[team] = []
                    rosters[team].append(player_name)
        return rosters

    def num_players(self):
        """Number of participants in this match."""
        return len(self.player_names())

    def num_teams(self):
        """Number of teams (rosters) in this match."""
        return len(self.rosters())

    def rankings(self, rank=None):
        """The rankings of each team from this match.

        Returns a map of rank : [players] for each team in the match.

        :param int rank: Get the specific rank number players for the match.
        """
        rankings = {}
        for event in self.events[::-1]:
            if event.event_type == "log_match_end":
                for player in event.characters:
                    ranking = player.ranking
                    if ranking not in rankings:
                        rankings[ranking] = []
                    rankings[ranking].append(player.name)
        if rank is not None:
            return rankings.get(rank, None)
        return rankings

    def winner(self):
        """The winner(s) of the match.

        Match winners as a list of player names.
        """
        return self.rankings(rank=1)

    @classmethod
    def from_json(cls, telemetry_json, pubg=None, url=None, shard=None):
        """Construct an instance of telemetry from the json response."""
        return cls(pubg, url, telemetry_json, shard)

    @classmethod
    def from_match_id(cls, match_id, pubg, shard=None):
        """Construct an instance of telemetry from a match id."""
        match = pubg.match(match_id, shard)
        url = match.telemetry_url()
        return cls(pubg, url, shard=shard)

    def map_name(self):
        """Get the map name for PC matches. None if not PC."""
        for event in self.events:
            if event.event_type == "log_match_start":
                map_id = getattr(event, "map_name", None)
                if map_id is not None:
                    return map_to_map_name[map_id]
                else:
                    return self._pubg.match(self.match_id()).map_name

    def map_id(self):
        """Get the map id for PC matches. None if not PC."""
        for event in self.events:
            if event.event_type == "log_match_start":
                map_id = getattr(event, "map_name", None)
                if map_id is not None:
                    return map_id
                else:
                    return self._pubg.match(self.match_id()).map_id

    def match_id(self):
        """The match id for the match."""
        for event in self.events:
            if event.event_type == "log_match_definition":
                return event.match_id

    def player_damages(self, include_pregame=False):
        """Get the player damages for the match.

        Returns a dict of attacker players as keys and values of damage
        attacker and victim positions with the values tuples. Each tuple has
        seven elements being (t, x_a, y_a, z_a, x_v, y_v, y_z) coordinates
        where a is attacker and v is victim.

        :param bool include_pregame: (default False) whether to include
            pre-game damage positions.
        """
        start = datetime.datetime.strptime(
            self.filter_by("log_match_start")[0].timestamp,
            "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        damages = {}
        attack_events = self.filter_by("log_player_attack")
        attackers = {}
        for event in attack_events:
            attackers[event.attack_id] = event.attacker

        damage_events = self.filter_by("log_player_take_damage")
        for event in damage_events:
            try:
                attacker = event.attacker.name
            except AttributeError:
                continue
            if attacker != "":
                timestamp = datetime.datetime.strptime(
                    event.timestamp, "%Y-%m-%dT%H:%M:%S.%fZ"
                )
                dt = (timestamp - start).total_seconds()
                if (not include_pregame and dt < 0) or event.attack_id == -1:
                    continue
                if attacker not in damages:
                    damages[attacker] = []
                attacker_location = attackers[event.attack_id].location
                damages[attacker].append(
                    (
                        dt,
                        attacker_location.x,
                        attacker_location.y,
                        attacker_location.z,
                        event.victim.location.x,
                        event.victim.location.y,
                        event.victim.location.z,
                    )
                )
        return damages

    def player_positions(self, include_pregame=False):
        """Get the player positions for the match.

        Returns a dict of players positions up to death with keys being player
        names and values being a list of tuples. Each tuple has four elements
        being (t, x, y, z) coordinates where t is taken from the event
        timestamps.

        :param bool include_pregame: (default False) whether to include
            pre-game player positions.
        """
        start = datetime.datetime.strptime(
            self.filter_by("log_match_start")[0].timestamp,
            "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        locations = self.filter_by("log_player_position")
        if not include_pregame:
            locations = [
                location for location in locations
                if location.elapsed_time > 0
            ]
        player_positions = {}
        dead = []
        for location in locations:
            timestamp = datetime.datetime.strptime(
                location.timestamp, "%Y-%m-%dT%H:%M:%S.%fZ"
            )
            dt = (timestamp - start).total_seconds()
            player = location.character.name
            if player not in player_positions:
                player_positions[player] = []
            elif player in dead:
                continue
            # (t, x, y, z)
            player_positions[player].append(
                (
                    dt,
                    location.character.location.x,
                    location.character.location.y,
                    location.character.location.z,
                )
            )
            if location.character.ranking > 1:
                dead.append(player)
        # cleanup
        for player, positions in player_positions.items():
            count = 0
            last = positions[-1]
            for pos in positions[::-1]:
                if pos == last:
                    count += 1
                else:
                    break
            if count > 0:
                player_positions[player] = positions[:-count]

        return player_positions

    def circle_positions(self):
        """Get the circle positions for the match.

        Returns a dict of circle positions with keys being circle colors and
        values being a list of tuples. Each tuple has five elements being
        (t, x, y, z, r) coordinates where t is taken from the "elapsedTime"
        field in the JSON response and r is the circle radius.

        The circle colors are "white", "blue", and "red"
        """
        game_states = self.filter_by("log_game_state_periodic")
        circle_positions = {
            "white": [],
            "blue": [],
            "red": [],
        }
        start = datetime.datetime.strptime(
            game_states[0].timestamp, "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        for game_state in game_states:
            timestamp = datetime.datetime.strptime(
                game_state.timestamp, "%Y-%m-%dT%H:%M:%S.%fZ"
            )
            dt = (timestamp - start).total_seconds()
            circle_positions["blue"].append(
                (
                    dt,
                    game_state.game_state.safety_zone_position.x,
                    game_state.game_state.safety_zone_position.y,
                    game_state.game_state.safety_zone_position.z,
                    game_state.game_state.safety_zone_radius,
                )
            )
            circle_positions["red"].append(
                (
                    dt,
                    game_state.game_state.red_zone_position.x,
                    game_state.game_state.red_zone_position.y,
                    game_state.game_state.red_zone_position.z,
                    game_state.game_state.red_zone_radius,
                )
            )
            circle_positions["white"].append(
                (
                    dt,
                    game_state.game_state.poison_gas_warning_position.x,
                    game_state.game_state.poison_gas_warning_position.y,
                    game_state.game_state.poison_gas_warning_position.z,
                    game_state.game_state.poison_gas_warning_radius,
                )
            )
        return circle_positions

    def care_package_positions(self, land=True):
        """Get the crate positions for the match.

        Returns the crate positions for a match as a list of tuples.
        Each tuple has four elements being
        (t, x, y, z) coordinates where t is taken from the "elapsedTime"
        field in the JSON response.
        """
        start = datetime.datetime.strptime(
            self.filter_by("log_match_start")[0].timestamp,
            "%Y-%m-%dT%H:%M:%S.%fZ"
        )

        if land:
            care_package_spawns = self.filter_by("log_care_package_land")
        else:
            care_package_spawns = self.filter_by("log_care_package_spawn")

        care_package_positions = []
        for care_package in care_package_spawns:
            package_time = datetime.datetime.strptime(
                care_package.timestamp, "%Y-%m-%dT%H:%M:%S.%fZ"
            )
            time_elapsed = (package_time - start).total_seconds()
            care_package_positions.append(
                (
                    time_elapsed,
                    care_package.item_package.location.x,
                    care_package.item_package.location.y,
                    care_package.item_package.location.z,
                )
            )
        return care_package_positions

    def match_length(self):
        """The length of the match in seconds."""
        for event in self.events[::-1]:
            elapsed_time = getattr(event, "elapsed_time", None)
            if elapsed_time is not None:
                return elapsed_time

    def started(self):
        """A timestamp of when the match started."""
        return self.events[0].timestamp

    def killed(self):
        """A list of player names of all killed players this match."""
        deaths = self.filter_by("log_player_kill")
        players_killed = []
        for death in deaths:
            players_killed.append(death.victim.name)
        # Telemetry data sometimes doesn't log all deaths
        # This is more reliable
        players = self.player_names()
        winner = self.winner()
        killed = set(players_killed) | (set(players) - set(winner))
        return list(killed)

    def playback_animation(self, filename="playback.html", **kwargs):
        """Generate a playback animation from the telemetry data.

        Generate an HTML5 animation using matplotlib and ffmpeg.
        Requires installation via ``pip install chicken-dinner[visual]``.

        :param filename: a file to generate for the animation (default
            "playback.html")
        :param bool labels: whether to label players by name
        :param int disable_labels_after: if passed, turns off player labels
            after number of seconds elapsed in game
        :param list label_players: a list of strings of player names that
            should be labeled
        :param bool dead_players: whether to mark dead players
        :param list dead_player_labels: a list of strings of players that
            should be labeled when dead
        :param bool zoom: whether to zoom with the circles through the playback
        :param float zoom_edge_buffer: how much to buffer the blue circle edge
            when zooming
        :param bool use_hi_res: whether to use the hi-res image, best to be set
            to True when using zoom
        :param bool color_teams: whether to color code different teams
        :param list highlight_teams: a list of strings of player names whose
            teams should be highlighted
        :param list highlight_players: a list of strings of player names who
            should be highlighted
        :param str highlight_color: a color to use for highlights
        :param bool highlight_winner: whether to highlight the winner(s)
        :param bool label_highlights: whether to label the highlights
        :param bool care_packages: whether to show care packages
        :param bool damage: whether to show PvP damage
        :param int end_frames: the number of extra end frames after game has
            been completed
        :param int size: the size of the resulting animation frame
        :param int dpi: the dpi to use when processing the animation
        :param bool interpolate: use linear interpolation to get frames with
            second-interval granularity
        :param int interval: interval between gameplay frames in seconds
        :param int fps: the frames per second for the animation
        """
        try:
            from chicken_dinner.visual.playback import create_playback_animation
        except ModuleNotFoundError as exc:
            print(
                "Use `pip install chicken_dinner[visual]` "
                "for visualization dependencies."
            )
            raise exc

        return create_playback_animation(self, filename, **kwargs)



"""Utility functions."""
import copy
import re


stats_map = {
    "DBNOs": "dbnos",
    "dBNOs": "dbnos",
    "top10s": "top_10s",
}


def camel_to_snake(name):
    try:
        return stats_map[name]
    except KeyError as exc:
        pass
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def remove_from_dict(d, keys):
    dcopy = copy.deepcopy(d)
    for k in keys:
        dcopy.pop(k, None)
    return dcopy



"""Shard samples model."""


class Samples(object):
    """Samples model.

    An object containing sample match data for a PUBG shard.

    :param pubg: an instance of the class :class:`chicken_dinner.pubgapi.PUBG`
    :param start: (optional) the timestamp from which samples are generated
    :param shard: the shard for the samples response
    """

    def __init__(self, pubg, start=None, shard=None):
        self._pubg = pubg
        self._shard = shard
        #: The start timestamp for this set of samples
        self.start = start
        #: The API response for this object
        self.response = self._pubg._core.samples(start, shard)

    @property
    def shard(self):
        """The shard for this player."""
        return self._shard or self._pubg.shard

    @property
    def data(self):
        """The data payload of the samples response."""
        return self.response["data"]

    @property
    def created_at(self):
        """The timestamp on which the set of samples was created."""
        return self.data["attributes"]["createdAt"]

    @property
    def id(self):
        """An id for the samples."""
        return self.data["id"]

    @property
    def match_ids(self):
        """A list of match ids from the samples response."""
        return [
            match["id"] for match in
            self.data["relationships"]["matches"]["data"]
        ]

    @property
    def title_id(self):
        """The title id associated with the samples."""
        return self.data["attributes"]["titleId"]

    @property
    def url(self):
        """A URL for this samples resource."""
        samples_url = SHARD_URL + self.shard + "/samples"
        if self.start is not None:
            return samples_url + "?filter[createdAt-start]=" + self.start
        else:
            return samples_url



"""Get the latest map images."""
import logging
import os

import requests


logger = logging.getLogger()
logger.setLevel("INFO")


MAPS_URL = "https://api.github.com/repos/pubg/api-assets/contents/Assets/Maps?ref=master"

MAP_ASSET_PATH = os.path.join(
    os.path.dirname(
        os.path.dirname(
            os.path.realpath(__file__)
        )
    ),
    "assets",
    "maps"
)


def update_maps(include_hi_res=True):
    """Update the maps local to this package."""
    response = requests.get(MAPS_URL).json()
    map_files = [f["download_url"] for f in response]

    for map_file in map_files:
        filename = map_file.split("/")[-1]
        logging.info("Downloading " + filename)

        response = requests.get(map_file)
        path = os.path.join(MAP_ASSET_PATH, filename)

        with open(path, "wb") as img:
            logging.info("Saving " + path)
            img.write(response.content)


if __name__ == "__main__":
    update_maps()



"""Function for generating playback animations."""
import json
import logging
import os
import random

import matplotlib.image as mpimg
import matplotlib.pyplot as plt
from matplotlib import patheffects
from matplotlib import rc
from matplotlib.animation import FuncAnimation
from matplotlib.offsetbox import AnnotationBbox
from matplotlib.offsetbox import OffsetImage

from chicken_dinner.constants import COLORS
from chicken_dinner.constants import map_dimensions


rc("animation", embed_limit=100)


MAP_ASSET_PATH = os.path.join(
    os.path.dirname(
        os.path.dirname(
            os.path.realpath(__file__)
        )
    ),
    "assets",
    "maps"
)


def create_playback_animation(
        telemetry,
        filename="playback.html",
        labels=True,
        disable_labels_after=None,
        label_players=[],
        dead_players=True,
        dead_player_labels=False,
        zoom=False,
        zoom_edge_buffer=0.5,
        use_hi_res=False,
        color_teams=True,
        highlight_teams=[],
        highlight_players=[],
        highlight_color="#FFFF00",
        highlight_winner=False,
        label_highlights=True,
        care_packages=True,
        damage=True,
        end_frames=20,
        size=5,
        dpi=100,
        interpolate=True,
        interval=1,
        fps=30,
    ):
    """Create a playback animation from telemetry data.

    Using matplotlib's animation library, create an HTML5 animation saved to
    disk relying on external ``ffmpeg`` library to create the video.

    To view the animation, open the resulting file in your browser.

    :param telemetry: an Telemetry instance
    :param filename: a file to generate for the animation (default
        "playback.html")
    :param bool labels: whether to label players by name
    :param int disable_labels_after: if passed, turns off player labels
        after number of seconds elapsed in game
    :param list label_players: a list of strings of player names that
        should be labeled
    :param bool dead_players: whether to mark dead players
    :param list dead_player_labels: a list of strings of players that
        should be labeled when dead
    :param bool zoom: whether to zoom with the circles through the playback
    :param float zoom_edge_buffer: how much to buffer the blue circle edge
        when zooming
    :param bool use_hi_res: whether to use the hi-res image, best to be set
        to True when using zoom
    :param bool color_teams: whether to color code different teams
    :param list highlight_teams: a list of strings of player names whose
        teams should be highlighted
    :param list highlight_players: a list of strings of player names who
        should be highlighted
    :param str highlight_color: a color to use for highlights
    :param bool highlight_winner: whether to highlight the winner(s)
    :param bool label_highlights: whether to label the highlights
    :param bool care_packages: whether to show care packages
    :param bool damage: whether to show PvP damage
    :param int end_frames: the number of extra end frames after game has
        been completed
    :param int size: the size of the resulting animation frame
    :param int dpi: the dpi to use when processing the animation
    :param bool interpolate: use linear interpolation to get frames with
        second-interval granularity
    :param int interval: interval between gameplay frames in seconds
    :param int fps: the frames per second for the animation
    """

    # Extract data
    positions = telemetry.player_positions()
    circles = telemetry.circle_positions()
    rankings = telemetry.rankings()
    winner = telemetry.winner()
    killed = telemetry.killed()
    rosters = telemetry.rosters()
    damages = telemetry.player_damages()
    package_spawns = telemetry.care_package_positions(land=False)
    package_lands = telemetry.care_package_positions(land=True)
    map_id = telemetry.map_id()
    mapx, mapy = map_dimensions[map_id]
    all_times = []
    for player, pos in positions.items():
        for p in pos:
            all_times.append(int(p[0]))
    all_times = sorted(list(set(all_times)))

    if highlight_winner:
        for player in winner:
            highlight_players.append(player)
        highlight_players = list(set(highlight_players))

    if highlight_teams is not None:
        for team_player in highlight_teams:
            for team_id, roster in rosters.items():
                if team_player in roster:
                    for player in roster:
                        highlight_players.append(player)
                    break

        highlight_players = list(set(highlight_players))

    if label_highlights:
        for player in highlight_players:
            label_players.append(player)
    label_players = list(set(label_players))

    team_colors = None
    if color_teams:
        # Randomly select colors from the pre-defined palette
        colors = COLORS
        idx = list(range(len(colors)))
        random.shuffle(idx)
        team_colors = {}
        count = 0
        for team_id, roster in rosters.items():
            for player in roster:
                team_colors[player] = colors[idx[count]]
            count += 1

    # Get the max "frame number"
    maxlength = 0
    for player, pos in positions.items():
        try:
            if pos[-1][0] > maxlength:
                maxlength = pos[-1][0]
        except IndexError:
            continue

    if interpolate:
        maxlength = max(all_times)
    else:
        maxlength = max([maxlength, len(circles)])

    # Initialize the plot and artist objects
    fig = plt.figure(frameon=False, dpi=dpi)
    ax = fig.add_axes([0, 0, 1, 1])
    ax.axis("off")

    if use_hi_res:
        if map_id == "Savage_Main":
            map_image = map_id + ".png"
        else:
            map_image = map_id + ".jpg"
    else:
        map_image = map_id + "_lowres.jpg"
    img_path = os.path.join(MAP_ASSET_PATH, map_image)
    try:
        img = mpimg.imread(img_path)
    except FileNotFoundError:
        raise FileNotFoundError(
            "High resolution images not included in package.\n"
            "Download images from https://github.com/pubg/api-assets/tree/master/Assets/Maps\n"
            "and place in folder: " + MAP_ASSET_PATH
        )
    implot = ax.imshow(img, extent=[0, mapx, 0, mapy])

    players = ax.scatter(-10000, -10000, marker="o", c="w", edgecolor="k", s=60, linewidths=1, zorder=20)
    deaths = ax.scatter(-10000, -10000, marker="X", c="r", edgecolor="k", s=60, linewidths=1, alpha=0.5, zorder=10)

    if highlight_players or highlight_teams:
        highlights = ax.scatter(-10000, -10000, marker="*", c=highlight_color, edgecolor="k", s=180, linewidths=1, zorder=25)
        highlights_deaths = ax.scatter(-10000, -10000, marker="X", c=highlight_color, edgecolor="k", s=60, linewidths=1, zorder=15)

    if labels:
        if label_players is not None:
            name_labels = {
                player_name: ax.text(0, 0, player_name, size=8, zorder=19)
                for player_name in positions if player_name in label_players
            }
        else:
            name_labels = {
                player_name: ax.text(0, 0, player_name, size=8, zorder=19)
                for player_name in positions
            }
        for label in name_labels.values():
            label.set_path_effects([patheffects.withStroke(linewidth=2, foreground="w")])

    blue_circle = plt.Circle((0, 0), 0, edgecolor="b", linewidth=2, fill=False, zorder=5)
    white_circle = plt.Circle((0, 0), 0, edgecolor="w", linewidth=2, fill=False, zorder=6)
    red_circle = plt.Circle((0, 0), 0, color="r", edgecolor=None, lw=0, fill=True, alpha=0.3, zorder=7)

    care_package_spawns, = ax.plot(-10000, -10000, marker="s", c="w", markerfacecoloralt="w", fillstyle="bottom", mec="k", markeredgewidth=0.5, markersize=10, lw=0, zorder=8)
    care_package_lands, = ax.plot(-10000, -10000, marker="s", c="r", markerfacecoloralt="b", fillstyle="bottom", mec="k", markeredgewidth=0.5, markersize=10, lw=0, zorder=9)

    damage_idx = 0
    damage_slots = 50
    damage_lines = []
    for k in range(damage_slots):
        dline, = ax.plot(-10000, -10000, marker="x", c="r", mec="r", markeredgewidth=5, markersize=10, lw=2, markevery=[1], alpha=0.5, zorder=50)
        damage_lines.append(dline)

    ax.add_patch(blue_circle)
    ax.add_patch(white_circle)
    ax.add_patch(red_circle)

    fig.subplots_adjust(left=0, right=1, bottom=0, top=1)
    fig.set_size_inches((size, size))

    ax.set_xlim([0, mapx])
    ax.set_ylim([0, mapy])

    # Frame init function
    def init():
        if labels:
            if highlight_players or highlight_teams:
                updates = players, deaths, highlights, highlights_deaths, blue_circle, red_circle, white_circle, *tuple(name_labels.values())
            else:
                updates = players, deaths, blue_circle, red_circle, white_circle, *tuple(name_labels.values())
        else:
            if highlight_players or highlight_teams:
                updates = players, deaths, highlights, highlights_deaths, blue_circle, red_circle, white_circle
            else:
                updates = players, deaths, blue_circle, red_circle, white_circle
        if care_packages:
            updates = *updates, care_package_lands, care_package_spawns
        if damage:
            updates = *updates, *damage_lines
        return updates

    def interpolate_coords(t, coords, tidx, vidx, step=False):
        inter = False
        for idx, coord in enumerate(coords):
            if coord[tidx] > t:
                inter = True
                break

        if not inter:
            return coords[-1][vidx]

        if idx == 0:
            return coords[0][vidx]
        else:
            v0 = coords[idx-1][vidx]
            t0 = coords[idx-1][tidx]

        v1 = coords[idx][vidx]
        t1 = coords[idx][tidx]

        if step:
            return v1
        else:
            return v0 + (t - t0) * (v1 - v0) / (t1 - t0)

    # Frame update function
    def update(frame):
        logging.info("Processing frame {frame}".format(frame=frame))
        try:
            if interpolate:
                blue_circle.center = (
                    interpolate_coords(frame, circles["blue"], 0, 1),
                    mapy - interpolate_coords(frame, circles["blue"], 0, 2))
                red_circle.center = (
                    interpolate_coords(frame, circles["red"], 0, 1, True),
                    mapy - interpolate_coords(frame, circles["red"], 0, 2, True))
                white_circle.center = (
                    interpolate_coords(frame, circles["white"], 0, 1, True),
                    mapy - interpolate_coords(frame, circles["white"], 0, 2, True))

                blue_circle.set_radius(
                    interpolate_coords(frame, circles["blue"], 0, 4))
                red_circle.set_radius(
                    interpolate_coords(frame, circles["red"], 0, 4, True))
                white_circle.set_radius(
                    interpolate_coords(frame, circles["white"], 0, 4, True))
            else:
                blue_circle.center = circles["blue"][frame][1], mapy - circles["blue"][frame][2]
                red_circle.center = circles["red"][frame][1], mapy - circles["red"][frame][2]
                white_circle.center = circles["white"][frame][1], mapy - circles["white"][frame][2]

                blue_circle.set_radius(circles["blue"][frame][4])
                red_circle.set_radius(circles["red"][frame][4])
                white_circle.set_radius(circles["white"][frame][4])
        except IndexError:
            pass

        xlim = ax.get_xlim()
        ylim = ax.get_ylim()

        xwidth = xlim[1] - xlim[0]
        ywidth = ylim[1] - ylim[0]

        if zoom:
            try:
                if interpolate:
                    margin_offset = (1 + zoom_edge_buffer) * interpolate_coords(frame, circles["blue"], 0, 4)
                    xmin = max([0, interpolate_coords(frame, circles["blue"], 0, 1) - margin_offset])
                    xmax = min([mapx, interpolate_coords(frame, circles["blue"], 0, 1) + margin_offset])
                    ymin = max([0, mapy - interpolate_coords(frame, circles["blue"], 0, 2) - margin_offset])
                    ymax = min([mapy, mapy - interpolate_coords(frame, circles["blue"], 0, 2) + margin_offset])
                else:
                    margin_offset = (1 + zoom_edge_buffer) * circles["blue"][frame][4]
                    xmin = max([0, circles["blue"][frame][1] - margin_offset])
                    xmax = min([mapx, circles["blue"][frame][1] + margin_offset])
                    ymin = max([0, mapy - circles["blue"][frame][2] - margin_offset])
                    ymax = min([mapy, mapy - circles["blue"][frame][2] + margin_offset])

                # ensure full space taken by map
                if xmax - xmin >= ymax - ymin:
                    if ymin == 0:
                        ymax = ymin + (xmax - xmin)
                    elif ymax == mapy:
                        ymin = ymax - (xmax - xmin)
                else:
                    if xmin == 0:
                        xmax = xmin + (ymax - ymin)
                    elif xmax == mapx:
                        xmin = xmax - (ymax - ymin)

                ax.set_xlim([xmin, xmax])
                ax.set_ylim([ymin, ymax])

                xwidth = xmax - xmin
                ywidth = ymax - ymin
            except IndexError:
                pass

        positions_x = []
        positions_y = []
        highlights_x = []
        highlights_y = []
        deaths_x = []
        deaths_y = []
        highlights_deaths_x = []
        highlights_deaths_y = []
        care_package_lands_x = []
        care_package_lands_y = []
        care_package_spawns_x = []
        care_package_spawns_y = []

        if color_teams:
            marker_colors = []
            death_marker_colors = []
        else:
            marker_colors = "w"
            death_marker_colors = "r"

        t = 0
        damage_count = 0
        for player, pos in positions.items():
            try:
                player_max = pos[-1][0]
                # This ensures the alive winner(s) stay on the map at the end.
                if frame >= player_max and player not in winner:
                    raise IndexError
                elif frame >= player_max and player not in killed:
                    fidx = frame if interpolate else -1
                else:
                    fidx = frame

                if interpolate:
                    t = max([t, fidx])
                else:
                    t = max([t, pos[fidx][0]])

                for package in package_spawns:
                    if package[0] < t and package[0] > t - 60:
                        care_package_spawns_x.append(package[1])
                        care_package_spawns_y.append(mapy - package[2])
                for package in package_lands:
                    if package[0] < t:
                        care_package_lands_x.append(package[1])
                        care_package_lands_y.append(mapy - package[2])

                # Update player positions
                if interpolate:
                    if fidx >= pos[-1][0] and player in killed:
                        raise IndexError
                    x = interpolate_coords(fidx, pos, 0, 1)
                    y = mapy - interpolate_coords(fidx, pos, 0, 2)
                else:
                    x = pos[fidx][1]
                    y = mapy - pos[fidx][2]

                # Update player highlights
                if player in highlight_players:
                    highlights_x.append(x)
                    highlights_y.append(y)
                else:
                    positions_x.append(x)
                    positions_y.append(y)
                    # Set colors
                    if color_teams:
                        marker_colors.append(team_colors[player])

                # Update labels
                if labels and player in label_players:
                    if disable_labels_after is not None and frame >= disable_labels_after:
                        name_labels[player].set_position((-100000, -100000))
                    else:
                        name_labels[player].set_position((x + 10000 * xwidth/mapx, y - 10000 * ywidth/mapy))

                # Update player damages
                if damage:
                    try:
                        for attack in damages[player]:
                            damage_frame = int(attack[0])
                            if damage_frame >= fidx + interval:
                                break
                            elif damage_frame >= fidx and damage_frame < fidx + interval:
                                damage_line_x = [attack[1], attack[4]]
                                damage_line_y = [mapy - attack[2], mapy - attack[5]]
                                damage_lines[damage_count].set_data(damage_line_x, damage_line_y)
                                damage_count += 1
                    except KeyError:
                        pass

            except IndexError as exc:
                # Set death markers
                if player in highlight_players:
                    highlights_deaths_x.append(pos[-1][1])
                    highlights_deaths_y.append(mapy - pos[-1][2])
                else:
                    deaths_x.append(pos[-1][1])
                    deaths_y.append(mapy - pos[-1][2])

                    # Set death marker colors
                    if color_teams:
                        death_marker_colors.append(team_colors[player])

                # Draw dead players names
                if labels and dead_player_labels and player in label_players:
                    name_labels[player].set_position((pos[-1][1] + 10000 * xwidth/mapx, mapy - pos[-1][2] - 10000 * ywidth/mapy))
                    name_labels[player].set_path_effects([patheffects.withStroke(linewidth=1, foreground="gray")])
                # Offscreen if labels are off
                elif labels and player in label_players:
                    name_labels[player].set_position((-100000, -100000))

        player_offsets = [(x, y) for x, y in zip(positions_x, positions_y)]
        if len(player_offsets) > 0:
            players.set_offsets(player_offsets)
        else:
            players.set_offsets([(x, y) for x, y in zip([-100000], [-100000])])

        if color_teams:
            players.set_facecolors(marker_colors)

        death_offsets = [(x, y) for x, y in zip(deaths_x, deaths_y)]
        if len(death_offsets) > 0:
            deaths.set_offsets(death_offsets)
        if color_teams:
            deaths.set_facecolors(death_marker_colors)

        if highlight_players is not None:
            highlight_offsets = [(x, y) for x, y in zip(highlights_x, highlights_y)]
            if len(highlight_offsets) > 0:
                highlights.set_offsets(highlight_offsets)

            highlight_death_offsets = [(x, y) for x, y in zip(highlights_deaths_x, highlights_deaths_y)]
            if len(highlight_death_offsets) > 0:
                highlights_deaths.set_offsets(highlight_death_offsets)

        if len(care_package_lands_x) > 0:
            care_package_lands.set_data(care_package_lands_x, care_package_lands_y)

        if len(care_package_spawns_x) > 0:
            care_package_spawns.set_data(care_package_spawns_x, care_package_spawns_y)

        # Remove the remaining slots
        for k in range(damage_count, damage_slots):
            damage_lines[k].set_data([], [])

        if labels:
            if highlight_players or highlight_teams:
                updates = players, deaths, highlights, highlights_deaths, blue_circle, red_circle, white_circle, *tuple(name_labels.values())
            else:
                updates = players, deaths, blue_circle, red_circle, white_circle, *tuple(name_labels.values())
        else:
            if highlight_players or highlight_teams:
                updates = players, deaths, highlights, highlights_deaths, blue_circle, red_circle, white_circle
            else:
                updates = players, deaths, blue_circle, red_circle, white_circle
        if care_packages:
            updates = *updates, care_package_lands, care_package_spawns
        if damage:
            updates = *updates, *damage_lines
        return updates

    # Create the animation
    animation = FuncAnimation(
        fig, update,
        frames=range(0, maxlength + end_frames, interval),
        interval=int(1000/fps),
        init_func=init, blit=True,
    )

    # Write the html5 to buffer
    h5 = animation.to_html5_video()

    # Save to disk
    logging.info("Saving file: {file}".format(file=filename))
    with open(filename, "w") as f:
        f.write(h5)
    logging.info("Saved file.")

    return True



"""Seasons model."""
from .season import Season


class Seasons(object):
    """Seasons model.

    An iterable containing instances of the class
    :class:`chicken_dinner.models.Season` for each season.

    :param pubg: an instance of the class :class:`chicken_dinner.pubgapi.PUBG`
    :param shard: the shard for the seasons response
    """

    def __init__(self, pubg, shard=None):
        self._pubg = pubg
        self._shard = shard
        #: The API response for this object
        self.response = self._pubg._core.seasons(self.shard)
        self._seasons = [Season(self._pubg, s, self.shard) for s in self.data]

    def __getitem__(self, idx):
        return self._seasons[idx]

    @property
    def shard(self):
        """The shard for this player."""
        return self._shard or self._pubg.shard

    @property
    def data(self):
        """The seasons' data payload from the response."""
        return self.response["data"]

    @property
    def ids(self):
        """Get a list of the season ids."""
        return [season.id for season in self._seasons]

    @property
    def url(self):
        """The url for the model response."""
        return self.response["links"]["self"]

    def current(self):
        """Get the current season.

        :return: a :class:`chicken_dinner.models.Season` instance for the
            current season
        """
        for season in self._seasons[::-1]:
            if season.is_current():
                return season
        return None



"""Setup."""
import io
from os import path
from setuptools import setup
from setuptools import find_packages


here = path.abspath(path.dirname(__file__))


# io.open for py27
with io.open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()


# import __version__ attributes
about = {}
with open(path.join(here, "chicken_dinner", "__version__.py")) as f:
    exec(f.read(), about)


setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    long_description=long_description,
    url=about["__url__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    license=about["__license__"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    keywords="pubg api wrapper playback animation chicken dinner",
    packages=find_packages(),
    install_requires=[
        "requests",
    ],
    extras_require={
        "visual": [
            "matplotlib",
            "pillow",
        ]
    },
    include_package_data=True,
    package_data={
        "": ["*lowres.jpg"]
    },
    zip_safe=False,
)



"""Match participant model."""
from chicken_dinner.util import camel_to_snake


class Participant(object):
    """Participant model.

    :param pubg: a PUBG instance
    :param match: the match object associated with this participant
    :param roster: the roster object associated with this participant
    :param data: the data payload from the participant response object
    :param str shard: the shard for the match for this participant
    """

    def __init__(self, pubg, match, roster, data, shard=None):
        self._pubg = pubg
        self._shard = shard
        #: The Match instance associated with this participant
        self.match = match
        #: The Roster instance associated with this participant
        self.roster = roster
        #: The data payload associated with this participant
        self.data = data
        #: Stats associated with this participant
        self.stats = {
            camel_to_snake(k): v
            for k, v in self.data["attributes"]["stats"].items()
        }

    @property
    def shard(self):
        """The shard for the match associated with this participant."""
        return self._shard or self._pubg.shard

    @property
    def id(self):
        """The player's account id."""
        return self.data["id"]

    @property
    def participant_id(self):
        """The match specific participant id of the player."""
        return self.data["id"]

    @property
    def player_id(self):
        """The player's account id."""
        return self.data["attributes"]["stats"]["playerId"]

    @property
    def name(self):
        """The player's in-game name."""
        return self.data["attributes"]["stats"]["name"]

    @property
    def actor(self):
        return self.data["attributes"]["actor"]

    @property
    def teammates(self):
        """A list of participant objects for this player's teammates."""
        return [
            p for p in self.roster.participants
            if p.participant_id != self.participant_id
        ]

    @property
    def teammates_player_names(self):
        """A list of player names for this player's teammates."""
        return [p.name for p in self.teammates]

    @property
    def teammates_player_ids(self):
        """A list of player account ids for this player's teammates."""
        return [p.player_id for p in self.teammates]

    @property
    def won(self):
        """Whether or not this player's team won the associated match."""
        return self.roster.won

    def get_player(self):
        """Get a Player object for this participant."""
        return self._pubg.player(self.player_id, self.shard)

    def get_player_season(self, season_id):
        """Get a PlayerSeason object for this participant.

        :param str season_id: the season id for which to retreive data
        """
        return self._pubg.player_season(self.player_id, season_id, self.shard)

    def get_current_player_season(self):
        """Get a PlayerSeason object for the current season."""
        return self.get_player_season("current")



from .player_season import PlayerSeason
from .player import Player
from .players import Players
from .samples import Samples
from .season import Season
from .seasons import Seasons
from .status import Status
from .tournament import Tournament
from .tournaments import Tournaments
